;;;  -*- mode: LISP; Package: CL-USER; Syntax: COMMON-LISP;  Base: 10 -*-
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; 
;;; Author      : Dan Bothell 
;;; Address     : Carnegie Mellon University
;;;             : Psychology Department
;;;             : Pittsburgh,PA 15213-3890
;;;             : db30+@andrew.cmu.edu
;;; 
;;; Copyright   : (c)2002-2004 Dan Bothell
;;; Availability: Covered by the GNU LGPL, see LGPL.txt
;;; 
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; 
;;; Filename    : dm-stepper-hack.lisp
;;; Version     : 1.5
;;; 
;;; Description : No system dependent code.
;;;             : This file contains the redefinitions of the 
;;;             : ACT-R functions that implement dm retrievals so that I
;;;             : can add my own hooks into them to record the details at
;;;             : the "right time" because by the time the chunk is 
;;;             : "retrieved" it's too late to get some of the info.
;;;             : It also contains the hook functions I'm using to record
;;;             : that info for the stepper.
;;; Bugs        : 
;;; 
;;; Todo        : Work this out with Christian so I don't have to hack it...
;;;             : Comment this mess.  
;;; ----- History -----
;;;
;;; 07/15/2002  Dan
;;;             : File creation 
;;; 09/25/2002  Dan
;;;             : Updated with the newest ACT-R code (v5.0.6)
;;; 10/01/2002  Dan
;;;             : Updated version to 1.1 and fixed the packaging
;;;             : for building a standalone in ACL.
;;; 4/22/2004   Dan [1.5]
;;;             : Added the license info.
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;(in-package :cl-user)

;#+(and :allegro-ide (not :ACTR-ENV-ALONE)) (in-package :cg-user)

(defvar *retrieval-record* nil)

(defvar *stepper-active* nil)


(defun retrieval-init ()
  (when *stepper-active*
    (setf *retrieval-record* (list (list (record-retrieval-request) 
                                         *activation-sources*)))))


(defun chunk-test-hook (chunk match-score)
  (when *stepper-active*
    
    (push-last (list chunk match-score) *retrieval-record*)))

(defvar *retrieval-init-hook-fn* #'retrieval-init)
(defvar *retrieval-chunk-test-hook-fn* #'chunk-test-hook)


(defun direct-test-and-bind (arguments)
  "Implements a direct match.  Tests for type and then slots, then bind
   variables.  Finally, call the function implementing the next clause.
   Arguments is a list of the wme's stack index, its type, a list of slots
   tests and bindings, and the next call if any.
   Returns t if a complete match is found, nil otherwise."
  (when *retrieval-init-hook-fn*
    (funcall *retrieval-init-hook-fn*))
  (let* ((wme-index (pop arguments))
         (wme (index-or-chunk wme-index))
         (wmetype (or (pop arguments) (when (wmep wme) (wme-type wme))))
         (slots-tested (pop arguments))
         (slots-bound (pop arguments)))
    
    (when *retrieval-chunk-test-hook-fn*
      (funcall *retrieval-chunk-test-hook-fn* wme nil))
    
    (if (and (wmep wme) (subtype (wme-type wme) wmetype))
      (let ((slots (wme-slots wme)))
        ;; Set bindings before testing slots
        (and (set-bindings slots slots-bound)
             (test-slots wme slots slots-tested) ; wme-index)
             (simulate-call arguments)))
      (signal-output *exact-matching-trace* "~S is not a CHUNK of type ~S."
                     wme wmetype))))

(defun indirect-test-and-bind (arguments)
  "Implements an indirect match.  For all subtypes of the wme, try all the wmes
   and test their slots, then bind variables.  Finally, call the function
   implementing the next clause.
   Arguments is a list of the wme's stack index, its type, a list of slots
   tests and bindings, and the next call if any.
   Returns t if a complete match is found, nil otherwise."
  (when *retrieval-init-hook-fn*
    (funcall *retrieval-init-hook-fn*))
  (let* ((wme-index (pop arguments))
         (wmetype (pop arguments))
         (slots nil)
         (slots-tested (pop arguments))
         (slots-bound (pop arguments))
         (retrieval-list nil))
    (signal-output *exact-matching-trace* "Matching CHUNKs of type ~S." wmetype)
    ;; Return the winning chunk or nil to handle-failure
    ;; If *enable-randomness* is on, then pick randomly otherwise pick the first
    (dolist (type (wme-type-subtypes wmetype)
                  (setf *retrieval-scheduler*
                        (if *enable-randomness*
                          (nth (random (length retrieval-list)) retrieval-list)
                          (first retrieval-list))))
      (dolist (wme (wme-type-wmes type))
        (setf slots (wme-slots wme))
        (when *retrieval-chunk-test-hook-fn*
          (funcall *retrieval-chunk-test-hook-fn* wme nil))

        (if wme-index
          (setf (instantiation-variable *instantiation* wme-index) wme)
          (setf *retrieval-scheduler* wme))
        ;; Set bindings before testing slots
        (when (and (set-bindings slots slots-bound)
                   (test-slots wme slots slots-tested)
                   (simulate-call arguments))
          ;; In action retrieval mode, keep the list of retrievals
          (unless wme-index
            (push-last *retrieval-scheduler* retrieval-list))
          ; pick and initialize a new instantiation
          (let* ((production (instantiation-production *instantiation*))
                 (next-instantiation (make-instantiation production))
                 (index 1)
                 (size (production-size production)))
            (loop
              (when (= index size) (return))
              (setf (instantiation-variable next-instantiation index)
                    (instantiation-variable *instantiation* index))
              (incf index 1))
            (setf *instantiation* next-instantiation)))))))


(defun direct-test-and-bind-pm (arguments)
  "Implements a direct match.  Tests for type and then slots, then bind
   variables.  Finally, call the function implementing the next clause.
   Arguments is a list of the wme's stack index, its type, a list of slots
   tests and bindings, and the next call if any.
   This is the rational analysis version, which handles partial matching too."
  (when *retrieval-init-hook-fn*
    (funcall *retrieval-init-hook-fn*))
  (let* ((wme-index (pop arguments))
         (wme (index-or-chunk wme-index))
         (wmetype (or (pop arguments) (when (wmep wme) (wme-type wme))))
         (slots-tested (pop arguments))
         (slots-bound (pop arguments)))
    (cond ((and (wmep wme) (subtype (wme-type wme) wmetype))
           (signal-output *activation-trace* "Sources of activation are: ~S" *activation-sources*)
           (let* ((slots (wme-slots wme))
                  ;; Set bindings before testing slots
                  (activation (and (set-bindings slots slots-bound)
                                   (if *partial-matching*
                                     (test-slots-pm wme slots slots-tested)
                                     (test-slots wme slots slots-tested)))))
             (setf *sum-exp-act* (exp (/ *retrieval-threshold* *temperature*)))
             ;; In exact match, only counts the chunk activation if it matches?
             (when activation
               (incf *sum-exp-act* (exp (/ activation *temperature*))))
             ;; If activation is below threshold, then reject it
             (when (and activation (< activation *retrieval-threshold*))
               (signal-output (if *partial-matching* *partial-matching-trace* *exact-matching-trace*)
                              "CHUNK ~S match score ~6,3F is below threshold ~6,3F: failure."
                              wme activation *retrieval-threshold*)
               (setf activation nil))
             (when *retrieval-chunk-test-hook-fn*
               (funcall *retrieval-chunk-test-hook-fn* wme activation))
             (add-latency wme (or activation *retrieval-threshold*))
             (when activation 
               (simulate-call arguments))))
          (t
           (when *retrieval-chunk-test-hook-fn*
             (funcall *retrieval-chunk-test-hook-fn* wme nil))
           (setf *sum-exp-act* (exp (/ *retrieval-threshold* *temperature*)))
           (add-latency wme *retrieval-threshold*)
           (signal-output (if *partial-matching* *partial-matching-trace* *exact-matching-trace*)
                          "~S is not a CHUNK of type ~S."
                          wme wmetype)))))

(defun indirect-test-and-bind-pm (arguments)
  "Implements an indirect match.  For all subtypes of the wme, try all the wmes
   and test their slots, then bind variables.  Finally, call the function
   implementing the next clause.
   Arguments is a list of the wme's stack index, its type, a list of slots
   tests and bindings, and the next call if any.
   This is the rational analysis version, which handles partial matching too.
   In addition, if blending is enabled then bind the consensus values."
  (when *retrieval-init-hook-fn*
    (funcall *retrieval-init-hook-fn*))
  (let* ((wme-index (pop arguments))
         (wmetype (pop arguments))
         (slots-tested (pop arguments))
         (slots-bound (pop arguments))
         (best-activation (if *blending* most-negative-short-float
                              *retrieval-threshold*))
         (best-wme nil)
         (slots nil)
         (activation nil)
         (temperature (when *blending* (or *temperature* (* (sqrt 2.0) *activation-noise*))))
         (total-prob (when *blending* (exp (/ *retrieval-threshold* temperature))))
         (chunk-probs nil))
    (signal-output *activation-trace* "Sources of activation are: ~S" *activation-sources*)
    (signal-output (if *partial-matching* *partial-matching-trace* *exact-matching-trace*)
                   "Matching CHUNKs of type ~S." wmetype)
    (setf *sum-exp-act* (exp (/ *retrieval-threshold* *temperature*)))
    ;; If blending is on, then do not apply retrieval threshold to individual chunks
    (let ((*retrieval-threshold* best-activation))
      (dolist (type (wme-type-subtypes wmetype))
        (dolist (wme (wme-type-wmes type))
          (setf slots (wme-slots wme))
          (if wme-index
            (setf (instantiation-variable *instantiation* wme-index) wme)
            (setf *retrieval-scheduler* wme))
          ;; Set bindings before testing slots
          (setf activation (and (set-bindings slots slots-bound)
                                (if *partial-matching*
                                  (test-slots-pm wme slots slots-tested)
                                  (test-slots wme slots slots-tested))))
          ;; In exact match, only counts the chunk activation if it matches?
          (when activation
            (incf *sum-exp-act* (exp (/ activation *temperature*))))
          (when (and activation *blending*)
            (let ((prob (exp (/ activation temperature))))
              (push (cons wme prob) chunk-probs)
              (incf total-prob prob)))
          (when (and activation (>= activation best-activation))
            (signal-output (if *partial-matching* *partial-matching-trace* *exact-matching-trace*)
                           "Activation ~6,3F is larger than previous best ~6,3F: selecting ~S."
                           activation best-activation wme)
            (setf best-activation activation)
            (setf best-wme wme))
          (when *retrieval-chunk-test-hook-fn*
            (funcall *retrieval-chunk-test-hook-fn* wme activation)))))
    (if (or (eq *blending* t)
            (and (eq *blending* 'rt)
                 (< best-activation *retrieval-threshold*)))
      (setf best-activation
            (set-blending-bindings chunk-probs slots-bound total-prob))
      (setf chunk-probs nil))
    (unless (or (eq *blending* 'rt)
                (>= best-activation *retrieval-threshold*))
      (setf best-activation *retrieval-threshold*)
      (setf best-wme nil))
    (add-latency (or best-wme 'failure) best-activation)
    (cond (best-wme
           (setf slots (wme-slots best-wme))
           (if wme-index
             (setf (instantiation-variable *instantiation* wme-index) best-wme)
             (setf *retrieval-scheduler* best-wme))
           (unless chunk-probs (set-bindings slots slots-bound))
           (simulate-call arguments))
          (t
           (signal-output (if *partial-matching* *partial-matching-trace* *exact-matching-trace*)
                          "No chunk reached activation threshold ~6,3F: matching fails."
                          *retrieval-threshold*)))))




(defun record-retrieval-request ()
  (let ((production (instantiation-production *instantiation*))
        (bindings nil))
      (dolist (binding (production-bindings production))
        (when (integerp (variable-index binding))
          (let ((value (instantiation-variable *instantiation* 
                                               (variable-index binding))))
            (when value
              (push (cons (variable-name binding) value) bindings)
              (push (cons (var-var> (variable-name binding))
                          (var-var> value))
                    bindings)))))
      
    (find '+retrieval> 
          (cdr (sublis bindings (production-text production))) :key #'car)))

#|
This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2.1 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this library; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
|#